#!/usr/bin/env pypy

import copy
import json
from pprint import pprint
import sys
from heapq import heappush, heappop
import signal
import game
import linemetric
import mutator
import power
import simplemetric
import wave

TAG = "kickass3"
DEPTH = 3
BREADTH = 4
APPLY = 1

# (2, 4)
# (4, 2)

def test(cand):
    return True

interim_solution=""
gseedno=0
gtaskid=0
class Node(object):
    def __init__(self, parent, board, move, met, queue_met, total_lines):
        self.parent = parent
        self.board = board
        self.met = met
        self.queue_met = queue_met
        self.total_lines = total_lines
        self.move = move
        self.level = 0 if parent is None else self.parent.level + 1

    def __cmp__(self, a):
        return cmp(self.queue_met, a.queue_met)
        # return cmp(-self.queue_met, -a.queue_met)
        # return cmp(self.met, a.met)
        # return cmp((-self.level, self.met), (-a.level, a.met))

    def get_parent_at_level(self, level):
        n = self
        while 1:
            if n.level == level:
                return n
            n = n.parent

    def get_full_move(self):
        n = self
        moves = []
        while n:
            moves.append(n.move)
            n = n.parent
        return ''.join(reversed(moves))

    def pprint(self):
        print "level=%d met=%.4f rel_move=%s" % (self.level, self.met, self.move)
        print self.board.show()


def find_optimal_node(root, rng, units, max_levels, options={},value=linemetric.value):
    max_depth = options.get('depth', DEPTH)
    max_breadth = options.get('breadth', BREADTH)
    y_coef = float(options.get('y_coef', 4.0))
    x_coef = float(options.get('x_coef', 1.0))
    x_coef_norm = options.get('x_coef_norm', 1)
    line_coef = float(options.get('line_coef', 100.0))
    level_coef = float(options.get('level_coef', 100.0))
    max_cands = 8
    # print max_depth, max_breadth

    x_coef = x_coef / root.board.w if x_coef_norm else x_coef
    if value == simplemetric.value:
        y_coef = 0.0
        x_coef = 0.0

    queue = []
    heappush(queue, root)
    max_level = root.level + max_depth

    cands = []  # sorted by node.met

    def add_cand_if_good(node):
        if not cands:
            cands.append(node)
        else:
            if len(cands) < max_cands:
                cands.append(node)
            else:
                if node.met < cands[-1].met:
                    cands[-1] = node
            cands.sort(key=lambda n: n.met)

    # max_idx = max_breadth ** max_depth
    # max_idx = (max_breadth * 0.7) ** max_depth
    # max_idx = 1.5 ** max_depth
    idx = 0
    waves_n = 0
    while queue:
        node = heappop(queue)
        unit = units[rng[node.level] % len(units)]
        if game.is_end_of_game(node.board, unit, node.level, max_levels):
            continue  # FIXME: check?

        wav = wave.SimpleWave(node.board, unit, *unit.spawn_coords)
        waves_n += 1

        leaves = []
        for x, y, z, locking_step in wav.get_unique_positions():
            tboard = copy.copy(node.board)
            # tboard.bmap = int(tboard.bmap)
            tboard.lock(unit.trans_bmap(x, y, z))
            # hsh = (hash(tboard.bmap) & 0xffffffffffff) * 2e-17
            hsh = z * 1e-14 + x * 1e-12 + y * 1e-10
            # hsh = 0
            # print tboard.show()
            move = wav.path_to(x, y, z) + locking_step

            # val = metric.value(tboard)
            # val = metric.value(tboard) + node.met          # score 31991
            # val = metric.value(tboard) - node.level * 10   # score 31975
            # val = metric.value(tboard) - node.level * 100  # score 36427
            # val = metric.value(tboard) - node.level * 1000 # score 37156
            # val = metric.value(tboard) - node.level * 5000   # score 37156
            # val = metric.value(tboard) - node.level * 20
            y_mass = y + unit.get_y_mass_center(z)
            x_mass = x + unit.get_x_mass_center(z)

            raw_met = value(tboard,y=y_mass,size=unit.span)
            lines2 = tboard.eliminate() ** 2
            total_lines = lines2 + node.total_lines


            met = hsh + raw_met
            q_met = hsh

            if value != simplemetric.value:
                met -= (total_lines * line_coef +
                        node.level * level_coef +
                        y_mass * y_coef +
                        x_mass * x_coef)
            else:
                met -= y_mass * 1e-5

            q_met = met - total_lines * 100  # - node.level*10
            # print met
            # q_met = 0

            leaves.append(Node(node, tboard, move, met, q_met, total_lines))

        leaves = sorted(leaves, key=lambda x: x.met)[:max_breadth]
        for leaf in leaves:
            idx += 1
            # leaf.pprint()
            # print leaf.met
            add_cand_if_good(leaf)
            if leaf.level < max_level:
                heappush(queue, leaf)
            # else:
            #     if best_metric is None or leaf.met <best_metric+0.1*abs(best_metric):
            #         cand_count+=1

        # if cand_count>cand_limit:
        #     break
        # if idx > max_idx:
        #     break

    # if best_node:
    #     best_node.pprint()
    # print idx, float(idx) / max_idx, waves_n, float(waves_n) / max_idx, best_metric
    return cands


def solve(g, options={}, pphrases=None):
    # TODO PERF: move initialization to parent function
    global interim_solution
    wop = game.WordsOfPower(pphrases or game.WORDS_OF_POWER)

    power_patterns = mutator.Patterns(g.units, wop.words)
    once_words = list(wop.words)
    once_words.sort(key=lambda w: len(w))

    board = copy.copy(g.board)
    rng = g.rng

    if len(g.units)==1 and g.units[0].size==1:
        value=simplemetric.value
    else:
        value=linemetric.value

    moves_list = []
    used_words = set()
    node = Node(None, board, '', value(board), 0, 0)
    while 1:
        sel_nodes = find_optimal_node(node, rng, g.units, g.source_length,
                                       options, value=value)
        if not sel_nodes:
            break

        best_metric = sel_nodes[0].met

        # same metric, except for hash. Higher tolerance values are bad.
        filt_nodes = [n for n in sel_nodes if n.met <= best_metric + 1e-7]
        if not filt_nodes: # may happen if metric is too large
            filt_nodes = [sel_nodes[0]]

        board = node.board
        unit_idx = rng[node.level] % len(g.units)
        unit = g.units[unit_idx]

        # do not use APPLY, it'll intersect with mutations
        # nlevel = min(node.level + APPLY, best_node.level)

        mut_res = []
        for cand_node in filt_nodes:
            next_node = cand_node.get_parent_at_level(node.level + 1)
            moves = next_node.move
            if moves:
                mt = mutator.Mutator(power_patterns.get_all_for_unit_idx(unit_idx),
                                     board, unit, moves[:-1])
                # mt.try_mutate_at(0, 'ei!')
                # print spam_words, once_words
                _, w = mt.put_one_then_spam(wop.filter(once_words, unit.symmetry),
                                            wop.spam[unit.symmetry])

                moves = mt.moves + next_node.move[-1]
                pscore, new_used = game.power_score_and_used(moves, wop.words,
                                                             used_words)
            else:
                pscore = 0
                new_used = set()
            mut_res.append({
                'node': next_node,
                'moves': moves,
                'pscore': pscore,
                'new_used': new_used
            })

        mut_res.sort(key=lambda r: (len(r['new_used']), r['pscore']))

        # uniq_res = set((len(r['new_used']), r['pscore']) for r in mut_res)
        # if len(uniq_res) > 1:
        #     print
        #     for r in mut_res:
        #         print len(r['new_used']), r['pscore']
        # print len(uniq_res)

        # pprint(mut_res)
        res = mut_res[-1]
        node = res['node']
        for w in res['new_used']:
            used_words.add(w)
            if w in once_words:
                once_words.remove(w)

            # print used_words
            # moves = node.move -- power off

            # print board.show()
            # print board.w, board.h
            # print node.move
            # print mt.moves

        moves_list.append(res['moves'])
        interim_solution += res['moves']
        # print '----------'
        # print 'best solution is:'
        # best_node.pprint()
        # print 'new root is:'
        # node.pprint()
        if node.level == g.source_length:
            break

    return ''.join(moves_list)


if __name__ == '__main__':
    params=sys.argv[1:]

    phrases=[]
    memlimit=0
    timelimit=0
    files=[]
    seedno=None
    for i in range(0,len(params),2):
        if params[i]=="-f":
            files.append(params[i+1])
        elif params[i]=="-p":
            phrases.append(params[i+1])
        elif params[i]=="-m":
            memlimit=float(params[i+1])
        elif params[i]=="-t":
            timelimit=float(params[i+1])
        elif params[i]=='-s':
            seedno=int(params[i+1])
        elif params[i]=='-c':
            pass
        else:
            raise Exception("Unknown parameter "+params[i])
    solutions = []        
    def avost(signum,frame):
        
        json.dump([{"problemId": gtaskid,
                    "seed": gseedno,
                    "tag": TAG,
                    "solution": interim_solution
                }],sys.stdout)
        sys.exit(0)

    signal.signal(signal.SIGINT,avost)
    signal.signal(signal.SIGTERM,avost)
    for infile in files:

        games = game.Games(infile)
        if seedno is None:
            gamestoplay=xrange(games.games_n())
        else:
            gamestoplay=[seedno]
        for i in gamestoplay:
            g = games.new_game(i)
            if g.board.w*g.board.h>2500:
                DEPTH=1
                BREADTH=8
                                             
            seed = g.rng.initial_seed()
            gseedno=seed
            gtaskid=games.task_id
            try:
                moves = solve(g,pphrases=phrases)
                solutions.append({
                    "problemId": games.task_id,
                    "seed": seed,
                    "tag": TAG,
                    "solution": moves
                })
            except:
                solutions.append({
                    "problemId": games.task_id,
                    "seed": seed,
                    "tag": TAG,
                    "solution": "ei! "
                })
                raise

    print json.dumps(solutions)
