#!/usr/bin/python

from game import Unit, MUTATION_TO_CMD, move_point, rotate


class PowerPattern(object):
    def __init__(self, unit, word):
        # TODO: doublecheck if move is possible
        self.units = []
        self.unit = unit
        for rot in xrange(unit.symmetry):
            pivot = unit.pivot
            members = [rotate(pivot, xy, rot) for xy in unit.members]
            points = set(members)
            for letter in word:
                cmd = MUTATION_TO_CMD[letter]
                members = [move_point(cmd, xy, pivot) for xy in members]
                pivot = move_point(cmd, pivot, pivot)
                for xy in members:
                    points.add(xy)

            points = list(points)  # just in case
            u = Unit(unit.pivot, points, unit.w, unit.h, _force_sym=1)
            self.units.append(u)

    def trans_bmap(self, dx, dy, rot):
        return self.units[rot % self.unit.symmetry].trans_bmap(dx, dy, 0)

    def trans_is_valid(self, dx, dy, rot):
        return self.units[rot % self.unit.symmetry].trans_is_valid(dx, dy, 0)


if __name__ == '__main__':
    import game
    import util

    units_defs = set()
    for fname in util.get_qualifiers():
        g = game.Games(fname)
        for u in g.units:
            mm = tuple(sorted(u.members))
            pivot = tuple(u.pivot)
        units_defs.add((pivot, mm))

    width = 36
    height = 60
    x0 = 10
    y0 = 6
    units = [Unit(p, m, width, height) for p, m in units_defs]

    for word in game.WORDS_OF_POWER:
        print '---------', word
        canonical = ''.join(MUTATION_TO_CMD[c] for c in word)
        for u in units:
            pp = PowerPattern(u, word)
            print '-' * 50
            print u.members
            for rot0 in range(u.symmetry):
                b_empty = game.Board(width, height)
                b_show = game.Board(width, height)
                b_ref = game.Board(width, height)
                b_test = game.Board(width, height)

                b_show.lock(u.trans_bmap(x0, y0, rot0))
                print b_show.show()

                x, y, rot = x0, y0, rot0
                b_ref.lock(u.trans_bmap(x, y, rot))
                for cmd in canonical:
                    if not b_empty.unit_fits(u, x, y, rot):
                        raise Exception("Increase board size!")
                    x, y, rot = game.forward_step(cmd, x, y, rot, u.symmetry)
                    b_ref.lock(u.trans_bmap(x, y, rot))
                print b_ref.show()

                b_test.lock(pp.trans_bmap(x0, y0, rot0))
                print b_test.show()

                if b_ref.bmap != b_test.bmap:
                    raise Exception("Invalid!")

                # raw_input()
