#!/usr/bin/env pypy
from multiprocessing import Process, Queue
import json
import random
import sys
import time

import game
import util
from lightningsolver import solve

def do_one(options={}, verbose=1):
    total_cmdlen = 0
    total_score = 0
    t00 = time.time()
    for fname in sorted(util.get_qualifiers())[:23]:
        games = game.Games(fname)
        t0 = time.time()
        moves = solve(games.new_game(0), options)
        dt = time.time() - t0

        moves_pwr = game.power_rewrite(moves)
        move_score, pwr_score = games.new_game(0).exec_and_score(moves_pwr)

        if verbose:
            print "%s %5d %6d %d %.2f" % (fname, len(moves),
                                          move_score, pwr_score, dt)

        total_cmdlen += len(moves)
        total_score += move_score + pwr_score

    total_time = time.time() - t00
    if verbose:
        print 'total:', total_cmdlen, total_score, total_time
    return total_score, total_time

def mp_wrapper(q, options):
    res = do_one(options, verbose=False)
    q.put((options, res))

if len(sys.argv) == 3:
    depth, breadth = map(int, sys.argv[1:])
    do_one({'depth': depth, 'breadth': breadth})
else:
    cores = int(sys.argv[1])
    params = [
        # (1, 2), (1, 3), (1, 4), (1, 5), (1, 6), (1, 7),
        # (2, 2), (2, 3), (2, 4), (2, 5), (2, 6),
        # (2, 2), (2, 3), (2, 4), (2, 5), (2, 6),
        # (3, 2), (3, 3), (3, 4),
        # (4, 2), (4, 3), (4, 4),
        # (5, 2), (5, 3), (8, 5)
        (1, 5), (2, 3), (2, 6), (3, 3), (3, 5), (4, 4)
    ]

    options = [{'depth': d, 'breadth': b, 'y_coef': y_coef, 'x_coef': x_coef,
                'x_coef_norm': x_coef_norm, 'line_coef': line_coef}
               for d, b in params
               for y_coef in [0.1, 0.5, 4.0, 10.0, 15.0]
               for x_coef in [1.0, 5.0, 10.0]
               for x_coef_norm in [1]
               for line_coef in [1.0, 10.0, 100.0, 1000.0]
    ]

    random.shuffle(options)
    working = 0
    q = Queue()
    all_res = []

    fmt = 'depth={depth} breadth={breadth} ' \
            'y_c={y_coef:8.2f} x_c={x_coef:8.2f} line_c={line_coef:7.1f}'
    def display(results):
        ss1, ss2 = [], []
        for opts, res in results:
            s_opt = fmt.format(**opts)
            s_res = '%7d %6.1fs' % (res[0], res[1])
            ss1.append(s_opt + ' ' + s_res)
            ss2.append(s_res + ' ' + s_opt)
        ss1.sort()
        ss2.sort()
        print '\n\n-----------'
        for s in ss1:
            print s
        print
        for s in ss2:
            print s

    while params or working:
        while working < cores and params:
            opts = options.pop(0)
            print "sub :", fmt.format(**opts)
            # print 'start', opts
            p = Process(target=mp_wrapper, args=(q, opts))
            p.start()
            working += 1
        opts, res = q.get()
        all_res.append((opts, res))
        with file('perf_results.json', 'a+') as f:
            print >>f, json.dumps((opts, res))
        working -= 1
        display(all_res)
        print
